#
# Magnifier Preferences - GnomeShell Magnifier settings via GSettings
#
# Copyright (C) 2011 Inclusive Design Research Centre
# Contributor: Joseph Scheuhammer <clown@alum.mit.edu>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2.1 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA


import dbus
from dbus.mainloop.glib import DBusGMainLoop

_bus = dbus.SessionBus(mainloop=DBusGMainLoop())

from gi.repository.Gio import Settings
import gtk
import gettext

SHOW_KEY                      = 'screen-magnifier-enabled';
SCREEN_POSITION_KEY           = 'screen-position';
MAG_FACTOR_KEY                = 'mag-factor';
LENS_MODE_KEY                 = 'lens-mode';
CLAMP_MODE_KEY                = 'scroll-at-edges';
MOUSE_MODE_KEY                = 'mouse-tracking';

INVERT_BRIGHTNESS_KEY         = 'invert-brightness';
BRIGHTNESS_KEY                = 'brightness';
CONTRAST_KEY                  = 'contrast';

CROSSHAIRS_SHOW_KEY           = 'show-cross-hairs';
CROSSHAIRS_CLIP_KEY           = 'cross-hairs-clip';
CROSSHAIRS_COLOR_KEY          = 'cross-hairs-color';
CROSSHAIRS_THICKNESS_KEY      = 'cross-hairs-thickness';
CROSSHAIRS_OPACITY_KEY        = 'cross-hairs-opacity';
CROSSHAIRS_LENGTH_KEY         = 'cross-hairs-length';


########################################################################
#                                                                      #
# D-Bus callbacks                                                      #
#                                                                      #
########################################################################

class DockedView:
    """For handling D-Bus calls to the magnifier first ZoomRegion asynchronously.
    """
    def __init__(self, magDbus, dockedViewPort):
        self.magDbus = magDbus
        self.viewPort = dockedViewPort
     
    def moveFirstZoomRegion(self):
        self.magDbus.getZoomRegions(reply_handler=self.gotZoomRegions,
                                    error_handler=self.getZoomRegionsErr)
    
    def gotZoomRegions(self, reply):
        if (reply and len(reply) > 0):
            self.doMoveZoomRegion(reply[0])
        else:
            print "gotZoomRegions(): No result from getZoomRegions"
    
    def doMoveZoomRegion(self, zoomerPath):
        if (zoomerPath == None):
            return
        try:
            zoomProxy = _bus.get_object('org.gnome.Magnifier', zoomerPath)
            zoomer = dbus.Interface(
                    zoomProxy, dbus_interface='org.gnome.Magnifier.ZoomRegion')
            zoomer.moveResize(self.viewPort, ignore_reply=True)
        except:
            print 'doMoveZoomRegion() failed'             
    
    def getZoomRegionsErr(self, error):
        logLine = 'getZoomRegions failed: ' + str(error)
        print logLine


########################################################################
#                                                                      #
# Utility for updating the brightness and contrast spinners.           #
#                                                                      #
########################################################################

def updateColorSpinners(newValue, rSpinner, gSpinner, bSpinner):
    gdkColor = gtk.gdk.color_parse(newValue)
    rSpinner.set_value((gdkColor.red >> 8)  / 255.0)
    gSpinner.set_value((gdkColor.green >> 8)  / 255.0)
    bSpinner.set_value((gdkColor.blue >> 8)  / 255.0)

########################################################################
#                                                                      #
# UI for magnifier preferences.                                        #
#                                                                      #
########################################################################

class MagPrefsUI:
    __gtype_name__ = "MagPrefsUI";

    def __init__(self):
        
        # Calculate the viewport for the 'docked' screen position.
        # note:  arbitrarily situated at left edge of screen, half way down;
        # 2/3 the width of the screen, and 1/3 the height.
        screenWidth = gtk.gdk.screen_width();
        screenHeight = gtk.gdk.screen_height();
        top = screenHeight / 3 * 2;
        height = screenHeight / 3;
        width = screenWidth / 2;
        self._dockedViewPort = [0, top, width, top + height];

        # Initialize DBus connection to the magnifier service.
        proxyObj = _bus.get_object("org.gnome.Magnifier",
                                    "/org/gnome/Magnifier")
        self.magnifier = dbus.Interface(proxyObj, "org.gnome.Magnifier")
        self.dockedView = DockedView(self.magnifier, self._dockedViewPort)

        # Initialize GSettings
        self.appSettings = Settings('org.gnome.desktop.a11y.applications')
        self.magSettings = Settings('org.gnome.desktop.a11y.magnifier')

        # Initialize state of the widgets in the UI.
        # CHANGE:  pass in <uiFile> instead of hard coding it?
        builder = gtk.Builder();
        builder.set_translation_domain(gettext.textdomain())
        builder.add_from_file('/usr/share/gs-mag/zoom-options.ui')
        
        # Set the widgets according to the current stored GSettings values.
        # Show/hide magnifier toggle ...
        self.magOnRadio = builder.get_object('magOnRadio');
        self.magOffRadio = builder.get_object('magOffRadio');
        aSetting = self.appSettings.get_boolean(SHOW_KEY);
        self.magOnRadio.set_active(aSetting);
        self.magOffRadio.set_active(not aSetting);
        self.appSettings.connect('changed::' + SHOW_KEY, \
            self.showHidePrefChanged)
        
        # ... mouse tracking mode radio buttons ...
        self._resetRadioNames(builder, \
            [ 'none', 'centered', 'proportional', 'push' ])
            
        self.mouseRadioGroup = builder.get_object('none').get_group()
        self._updateMouseRadioGroup(self.magSettings, MOUSE_MODE_KEY)
        self.magSettings.connect('changed::' + MOUSE_MODE_KEY, \
            self._updateMouseRadioGroup)
        
        # ... screen positions ...
        self._resetRadioNames(builder, \
            [ 'full-screen', 'top-half', 'bottom-half', 'left-half', 'right-half', 'docked' ])
        
        self._dockedRadio = builder.get_object('docked')
        self.screenRadioGroup = self._dockedRadio.get_group()
        self._updateScreenRadioGroup(self.magSettings, SCREEN_POSITION_KEY)
        self.magSettings.connect('changed::' + SCREEN_POSITION_KEY, \
            self._updateScreenRadioGroup)

        # ... magnification factor slider ...
        magSlider = builder.get_object('magFactorSlider')
        aSetting = self.magSettings.get_double(MAG_FACTOR_KEY)
        magSlider.set_value(aSetting)
        magAdjustment = magSlider.get_adjustment()
        # 0 = G_SETTINGS_BIND_DEFAULT
        self.magSettings.bind(MAG_FACTOR_KEY, magAdjustment, 'value', 0)
                
        # ... lens mode checkbox ...
        lensMode = builder.get_object('moveableLensCheckbox')
        aSetting = self.magSettings.get_boolean(LENS_MODE_KEY)
        lensMode.set_active(aSetting)
        self.magSettings.bind(LENS_MODE_KEY, lensMode, 'active', 0)
        
        # ... clamp scrolling at screen edges checkbox ...
        edgeScrollMode = builder.get_object('scrollAtEdges');
        aSetting = self.magSettings.get_boolean(CLAMP_MODE_KEY);
        edgeScrollMode.set_active(aSetting);
        self.magSettings.bind(CLAMP_MODE_KEY, edgeScrollMode, 'active', 0)
        
        # ... cross hairs: show/hide ...
        showCrosshairsMode = builder.get_object('xHairsShowHide');
        aSetting = self.magSettings.get_boolean(CROSSHAIRS_SHOW_KEY);
        showCrosshairsMode.set_active(aSetting);
        self.magSettings.bind(CROSSHAIRS_SHOW_KEY, showCrosshairsMode, \
            'active', 0)
        
        # ... cross hairs: color ...
        self.xhairsColor = builder.get_object('xHairsPicker');
        aSetting = self.magSettings.get_string(CROSSHAIRS_COLOR_KEY);
        gdkColor = gtk.gdk.color_parse(aSetting);
        self.xhairsColor.set_color(gdkColor);
        self.magSettings.connect('changed::' + CROSSHAIRS_COLOR_KEY,\
            self.xhairsColorPrefChanged)

        # ... cross hairs: opacity ...
        crosshairsOpacity = builder.get_object('xHairsOpacitySlider')
        aSetting = self.magSettings.get_double(CROSSHAIRS_OPACITY_KEY)
        crosshairsOpacity.set_value(aSetting)
        xHairsOpacityAdjustment = crosshairsOpacity.get_adjustment()
        self.magSettings.bind(CROSSHAIRS_OPACITY_KEY, xHairsOpacityAdjustment, \
            'value', 0)
        
        # ... cross hairs: thickness ...
        crosshairsThickness = builder.get_object('xHairsThicknessSpinner')
        aSetting = self.magSettings.get_int(CROSSHAIRS_THICKNESS_KEY)
        crosshairsThickness.set_value(aSetting)
        self.magSettings.bind(CROSSHAIRS_THICKNESS_KEY, crosshairsThickness, \
            'value', 0)
        
        # ... cross hairs: clip ...
        clipCrosshairsMode = builder.get_object('xHairsClipCheckbox');
        aSetting = self.magSettings.get_boolean(CROSSHAIRS_CLIP_KEY);
        clipCrosshairsMode.set_active(aSetting);
        self.magSettings.bind(CROSSHAIRS_CLIP_KEY, clipCrosshairsMode, \
            'active', 0)
        
        # ... Cross hairs: length
        crosshairsLength = builder.get_object('xHairsLengthSlider');
        aSetting = self.magSettings.get_int(CROSSHAIRS_LENGTH_KEY);
        crosshairsLength.set_value(aSetting);
        xhairsLengthAdjustment = crosshairsLength.get_adjustment()
        self.magSettings.bind(CROSSHAIRS_LENGTH_KEY, xhairsLengthAdjustment, \
            'value', 0)       

        # ... Brightness ...
        invertBrightness = builder.get_object('invertBrightnessCheckbox')
        aSetting = self.magSettings.get_boolean(INVERT_BRIGHTNESS_KEY)
        invertBrightness.set_active(aSetting)
        self.magSettings.bind(INVERT_BRIGHTNESS_KEY, invertBrightness, \
            'active', 0)

        self.brightRed = builder.get_object('brightRedSpinner');
        self.brightGreen = builder.get_object('brightGreenSpinner')
        self.brightBlue = builder.get_object('brightBlueSpinner')
        aSetting = self.magSettings.get_string(BRIGHTNESS_KEY)
        updateColorSpinners(aSetting, \
            self.brightRed, self.brightGreen, self.brightBlue)
        self.magSettings.connect('changed::' + BRIGHTNESS_KEY, \
            self._updateBrightnessSpinners)

        # ... Contrast.
        self.contrastRed = builder.get_object('contrastRedSpinner')
        self.contrastGreen = builder.get_object('contrastGreenSpinner')
        self.contrastBlue = builder.get_object('contrastBlueSpinner')
        aSetting = self.magSettings.get_string(CONTRAST_KEY)
        updateColorSpinners(aSetting, \
            self.contrastRed, self.contrastGreen, self.contrastBlue)
        self.magSettings.connect('changed::' + CONTRAST_KEY, \
            self._updateContrastSpinners)
       
        # Attach widget handlers
        instance_attributes = {}
        for attribute in dir(self.__class__):
            instance_attributes[attribute] = getattr(self, attribute)
        builder.connect_signals(instance_attributes)

        # Handle closing.
        mainWindow = builder.get_object("magPrefsDialog");
        mainWindow.connect("destroy", gtk.main_quit);
        mainWindow.connect("delete_event", lambda window, event: False);
        closeButton = builder.get_object("closeButton");
        closeButton.connect("clicked", lambda button: mainWindow.destroy())

        mainWindow.show_all()
    
    def showHidePrefChanged(self, settings, key):
        toggle = settings.get_boolean(SHOW_KEY)
        self.magOnRadio.set_active(toggle)
        self.magOffRadio.set_active(not toggle)
    
    def magFactorChanged(self, settings, key):
        self.magSlider.set_value(settings.get_double(MAG_FACTOR_KEY))
       
    def doShowHide(self, radioButton):
        self.appSettings.set_boolean(SHOW_KEY, radioButton.get_active())
    
    def doMouseTracking(self, radioButton):
        if radioButton.get_active():
            self.magSettings.set_string(MOUSE_MODE_KEY, radioButton.get_name())
    
    def doScreenPosition(self, radioButton):
        if radioButton.get_active():
            if radioButton == self._dockedRadio:
                self.dockedView.moveFirstZoomRegion()
            else:
                self.magSettings.set_string(SCREEN_POSITION_KEY, \
                    radioButton.get_name())
    
    def doMagnification(self, slider):
        # TODO: two decimal places
        magFactor = slider.get_value()
        self.magSettings.set_double(MAG_FACTOR_KEY, magFactor)
    
    def _updateMouseRadioGroup(self, settings, key):
        self._updateRadioGroup(settings, key, self.mouseRadioGroup)
    
    def _updateScreenRadioGroup(self, settings, key):
        self._updateRadioGroup(settings, key, self.screenRadioGroup)
    
    def xhairsColorPrefChanged(self, settings, key):
        color = gtk.gdk.color_parse(settings.get_string(CROSSHAIRS_COLOR_KEY))
        self.xhairsColor.set_color(color)
    
    def doCrosshairsColor(self, colorButton):
        # colorVal is in the form '#rrrrggggbbbb'.  Convert to '#rrggbb'.
        colorVal = colorButton.get_color().to_string()
        cVal = '#'
        cVal += colorVal[1:3]   # red
        cVal += colorVal[5:7]   # green
        cVal += colorVal[9:11]  # blue
        self.magSettings.set_string(CROSSHAIRS_COLOR_KEY, cVal)
    
    def _updateRadioGroup(self, settings, key, radioGroup):
        value = settings.get_string(key)
        for aRadio in radioGroup:
            radioName = aRadio.get_name()
            if radioName == value:
                aRadio.set_active(True)
            else:
                aRadio.set_active(False)
    
    # Hack to get names back into the radio buttons.  The names *are* defined
    # in the .ui file, but python loses that info (?)
    def _resetRadioNames(self, builder, namesArray):
        for aName in namesArray:
            aRadio = builder.get_object(aName)
            if aRadio:
                aRadio.set_name(aName)
    
    def doBrightness(self, spinner):
        red = self.brightRed.get_value() * 255
        green = self.brightGreen.get_value() * 255
        blue = self.brightBlue.get_value() * 255
        bString = '#' +'%(R)02x%(G)02x%(B)02x' % {"R":red, "G": green, "B": blue}
        self.magSettings.set_string(BRIGHTNESS_KEY, bString);

    def doContrast(self, spinner):
        red = self.contrastRed.get_value() * 255
        green = self.contrastGreen.get_value() * 255
        blue = self.contrastBlue.get_value() * 255
        cString = '#' +'%(R)02x%(G)02x%(B)02x' % {"R":red, "G": green, "B": blue}
        self.magSettings.set_string(CONTRAST_KEY, cString);

    def _updateBrightnessSpinners(self, settings, key):
        newBrightness = self.magSettings.get_string(BRIGHTNESS_KEY)
        updateColorSpinners(newBrightness, \
            self.brightRed, self.brightGreen, self.brightBlue)
    
    def _updateContrastSpinners(self, settings, key):
        newContrast = self.magSettings.get_string(CONTRAST_KEY)
        updateColorSpinners(newContrast, \
            self.contrastRed, self.contrastGreen, self.contrastBlue)

# launch it
magPrefsUI = MagPrefsUI();
gtk.main();
